/*
   https://howtodoinjava.com/java7/forkjoin-framework-tutorial-forkjoinpool-example/
*/
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.RecursiveTask;

import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.TimeUnit;

public class FolderProcessor extends RecursiveTask<List<String>>
{
   private static final long serialVersionUID = 1L;
   private final String path;
   private final String extension;

   public FolderProcessor(String path, String extension)
   {
      this.path = path;
      this.extension = extension;
   }


   @Override
   protected List<String> compute()
   {
      // Names of the files stored in the folder.
      List<String> list = new ArrayList<String>();

      // FolderProcessor tasks that process the subfolders stored in the folder.
      List<FolderProcessor> tasks = new ArrayList<FolderProcessor>();

      File file = new File(path);
      File content[] = file.listFiles();
      // For each element in the folder, if there is a subfolder,
      // create a new FolderProcessor object and execute it
      // asynchronously using the fork() method.
      if (content != null)
      {
         for (int i = 0; i < content.length; i++)
         {
            if (content[i].isDirectory())
            {
               FolderProcessor task = new FolderProcessor(content[i].getAbsolutePath(), extension);
               task.fork();
               tasks.add(task);
            }
            else
            {
               if (checkFile(content[i].getName()))
               {
                  list.add(content[i].getAbsolutePath());
               }
            }
         }
      }

      if ( tasks.size() > 0 )
      {
         System.out.printf("%s: %d tasks ran.\n", file.getAbsolutePath(), tasks.size());
      }

      for (FolderProcessor item : tasks)
      {
         list.addAll(item.join());
      }
      return list;
   }


   // This method compares if the name of a file passed as a
   // parameter ends with the extension you are looking for.
   private boolean checkFile(String name)
   {
      return name.endsWith(extension);
   }


   public static void main(String[] args)
   {
      // Create three FolderProcessor tasks. Initialize each one with a different folder path.
      FolderProcessor apps      = new FolderProcessor("C:\\Program Files", "log");
      FolderProcessor system    = new FolderProcessor("C:\\Windows\\System32", "dll");
      FolderProcessor documents = new FolderProcessor("C:\\Documents And Settings", "txt");

      // Put each task into the fork/join work queue.
      ForkJoinPool.commonPool().invoke(apps);
      ForkJoinPool.commonPool().invoke(system);
      ForkJoinPool.commonPool().invoke(documents);

      // Write to the console information about the status of the pool every
      // second until the three tasks have finished their execution.
      do
      {
         System.out.printf("******************************************\n");
         System.out.printf("Main: Parallelism: %d\n",
                            ForkJoinPool.commonPool().getParallelism());
         System.out.printf("Main: Active Threads: %d\n",
                            ForkJoinPool.commonPool().getActiveThreadCount());
         System.out.printf("Main: Task Count: %d\n",
                            ForkJoinPool.commonPool().getQueuedTaskCount());
         System.out.printf("Main: Steal Count: %d\n",
                            ForkJoinPool.commonPool().getStealCount());
         System.out.printf("******************************************\n");
         try
         {
            TimeUnit.SECONDS.sleep(1);
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
      } while ( !system.isDone() || !apps.isDone() || !documents.isDone() );

      List<String> results1 = apps.join();
      List<String> results2 = system.join();
      List<String> results3 = documents.join();

      System.out.printf("Apps: %d files found.\n",      results1.size());
      System.out.printf("System: %d files found.\n",    results2.size());
      System.out.printf("Documents: %d files found.\n", results3.size());
   }
}
